# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals


__all__ = [
    'ext',
    '__version__',
]

# http://stackoverflow.com/a/2073599/
from pkg_resources import require
__version__ = require('filters')[0].version
del require

# Make the base filters accessible from the top level of the package.
# Note that the order is important here, due to dependencies.
from .base import *
from .handlers import *
from .number import *
from .simple import *
from .complex import *
from .string import *

# Additional filters are loaded into a separate namespace, so that IDEs
# don't go insane.
from filters.extensions import FilterExtensionRegistry
ext = FilterExtensionRegistry()
del FilterExtensionRegistry



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from collections import OrderedDict
from typing import Any, Dict, Generator, Iterable, Mapping, Optional, \
    Text, Tuple, Union

from six import iteritems, iterkeys, python_2_unicode_compatible

from filters.base import BaseFilter, FilterCompatible, FilterError, Type
from filters.string import Unicode

__all__ = [
    'FilterMapper',
    'FilterRepeater',
]


@python_2_unicode_compatible
class FilterRepeater(BaseFilter):
    """
    Applies a filter to every value in an Iterable.

    You can apply a FilterRepeater to a dict (or other Mapping).  The
    filters will be applied to the Mapping's values.

    Note:  The resulting value will be coerced to a list or OrderedDict
    (depending on the input value).
    """
    CODE_EXTRA_KEY = 'unexpected'

    templates = {
        CODE_EXTRA_KEY: 'Unexpected key "{key}".',
    }

    mapping_result_type     = OrderedDict
    sequence_result_type    = list

    def __init__(self, filter_chain, restrict_keys=None):
        # type: (FilterCompatible, Optional[Iterable]) -> None
        """
        :param filter_chain:
            The filter(s) that will be applied to each item in the
            incoming iterables.

        :param restrict_keys:
            Only these keys/indexes will be allowed (any other
            keys/indexes encountered will be treated as invalid
            values).

            Important:  If this is an empty container will result in
            EVERY key/index being rejected!

            Set to ``None`` (default) to allow any key/index.
        """
        super(FilterRepeater, self).__init__()

        self._filter_chain = self.resolve_filter(filter_chain, parent=self)

        self.restrict_keys = (
            None
                if restrict_keys is None
                else set(restrict_keys)
        )

    def __str__(self):
        return '{type}({filter_chain})'.format(
            type            = type(self).__name__,
            filter_chain    = self._filter_chain,
        )

    # noinspection PyProtectedMember
    @classmethod
    def __copy__(cls, the_filter):
        # type: (FilterRepeater) -> FilterRepeater
        """
        Creates a shallow copy of the object.
        """
        new_filter = super(FilterRepeater, cls).__copy__(the_filter) # type: FilterRepeater

        new_filter._filter_chain = the_filter._filter_chain
        new_filter.restrict_keys = the_filter.restrict_keys

        return new_filter

    def _apply(self, value):
        value = self._filter(value, Type(Iterable)) # type: Iterable

        if self._has_errors:
            return None

        result_type = (
            self.mapping_result_type
                if isinstance(value, Mapping)
                else self.sequence_result_type
        )

        return result_type(self.iter(value))

    def iter(self, value):
        # type: (Iterable) -> Generator[Any]
        """
        Iterator version of :py:meth:`apply`.
        """
        if value is not None:
            if isinstance(value, Mapping):
                for k, v in iteritems(value):
                    u_key = self.unicodify_key(k)

                    if (
                            (self.restrict_keys is None)
                        or  (k in self.restrict_keys)
                    ):
                        yield k, self._apply_item(u_key, v, self._filter_chain)
                    else:
                        # For consistency with FilterMapper, invalid
                        # keys are not included in the filtered
                        # value (hence this statement does not
                        # ``yield``).
                        self._invalid_value(
                            value   = v,
                            reason  = self.CODE_EXTRA_KEY,
                            sub_key = u_key,
                        )
            else:
                for i, v in enumerate(value):
                    u_key = self.unicodify_key(i)

                    if (
                            (self.restrict_keys is None)
                        or  (i in self.restrict_keys)
                    ):
                        yield self._apply_item(u_key, v, self._filter_chain)
                    else:
                        # Unlike in mappings, it is not possible to
                        # identify a "missing" item in a collection,
                        # so we have to ensure that something ends up
                        # in the filtered value at the same position
                        # as the invalid incoming value.
                        yield self._invalid_value(
                            value   = v,
                            reason  = self.CODE_EXTRA_KEY,
                            sub_key = u_key,
                        )

    def _apply_item(self, key, value, filter_chain):
        # type: (Text, Any, FilterCompatible) -> Any
        """
        Applies filters to a single value in the iterable.

        Override this method in a subclass if you want to customize the
        way specific items get filtered.
        """
        return self._filter(value, filter_chain, sub_key=key)

    @staticmethod
    def unicodify_key(key):
        # type: (Any) -> Text
        """
        Converts a key value into a unicode so that it can be
        represented in e.g., error message contexts.
        """
        if key is None:
            return 'None'

        try:
            return Unicode().apply(key)
        except FilterError:
            return repr(key)


@python_2_unicode_compatible
class FilterMapper(BaseFilter):
    """
    Given a dict of filters, applies each filter to the corresponding
    value in incoming mappings.

    The resulting value is an OrderedDict.  The order of keys in the
    ``filter_map`` passed to the initializer determines the order of
    keys in the filtered value.

    Note: The order of extra keys is undefined, but they will always be
    last.
    """
    CODE_EXTRA_KEY      = 'unexpected'
    CODE_MISSING_KEY    = 'missing'

    templates = {
        CODE_EXTRA_KEY:     'Unexpected key "{actual_key}".',
        CODE_MISSING_KEY:   '{key} is required.',
    }

    def __init__(
            self,
            filter_map,
            allow_missing_keys  = True,
            allow_extra_keys    = True,
    ):
        # type: (Dict[Text, FilterCompatible], Union[bool, Iterable[Text]], Union[bool, Iterable[Text]]) -> None
        """
        :param filter_map:
            This mapping also determines the key order of the resulting
            OrderedDict.  If necessary, make sure that your code
            provides ``filter_map`` as an OrderedDict.

        :param allow_missing_keys:
            Determines how values with missing keys (according to
            ``filter_map``) get handled:

            - True: The missing values are set to ``None`` and then
              filtered as normal.
            - False: Missing keys are treated as invalid values.
            - <Iterable>: Only the specified keys are allowed to be
              omitted.

        :param allow_extra_keys:
            Determines how values with extra keys (according to
            ``filter_map``) get handled:

            - True: The extra values are passed through to the filtered
              value.
            - False: Extra values are treated as invalid values and
              omitted from the filtered value.
            - <Iterable>: Only the specified extra keys are allowed.
        """
        super(FilterMapper, self).__init__()

        self._filters = OrderedDict()

        self.allow_missing_keys = (
            set(allow_missing_keys)
                if isinstance(allow_missing_keys, Iterable)
                else bool(allow_missing_keys)
        )

        self.allow_extra_keys = (
            set(allow_extra_keys)
                if isinstance(allow_extra_keys, Iterable)
                else bool(allow_extra_keys)
        )

        if filter_map:
            for key, filter_chain in iteritems(filter_map): # type: Tuple[Text, BaseFilter]
                #
                # Note that the normalized Filter could be `None`.
                #
                # This has the effect of making a key "required"
                # (depending on `allow_missing_keys`) without
                # applying any Filters to the value.
                #
                self._filters[key] =\
                    self.resolve_filter(filter_chain, parent=self, key=key)

        # If the filter map is an OrderedDict, we should try to
        # preserve order when applying the filter.  Otherwise use a
        # plain ol' dict to improve readability.
        self.result_type = (
            OrderedDict
                if isinstance(filter_map, OrderedDict)
                else dict
        )


    def __str__(self):
        return '{type}({filters})'.format(
            type    = type(self).__name__,
            filters = ', '.join(
                '{key}={filter}'.format(key=key, filter=filter_chain)
                    for key, filter_chain in iteritems(self._filters)
            ),
        )

    def _apply(self, value):
        value = self._filter(value, Type(Mapping)) # type: Mapping

        if self._has_errors:
            return None

        return self.result_type(self.iter(value))

    def iter(self, value):
        # type: (Mapping) -> Generator[Text, Any]
        """
        Iterator version of :py:meth:`apply`.
        """
        if value is not None:
            # Apply filtered values first.
            for key, filter_chain in iteritems(self._filters):
                if key in value:
                    yield key, self._apply_item(key, value[key], filter_chain)

                elif self._missing_key_allowed(key):
                    # Filter the missing value as if it was set to
                    # ``None``.
                    yield key, self._apply_item(key, None, filter_chain)

                else:
                    # Treat the missing value as invalid.
                    yield key, self._invalid_value(
                        value   = None,
                        reason  = self.CODE_MISSING_KEY,
                        sub_key = key,
                    )

            # Extra values go last.
            # Note that we iterate in sorted order, in case the result
            # type preserves ordering.
            # https://github.com/eflglobal/filters/issues/13
            for key in sorted(
                            set(iterkeys(value))
                        -   set(iterkeys(self._filters))
            ):
                if self._extra_key_allowed(key):
                    yield key, value[key]
                else:
                    unicode_key = self.unicodify_key(key)

                    # Handle the extra value just like any other
                    # invalid value, but do not include it in the
                    # result (note that there is no ``yield`` here).
                    self._invalid_value(
                        value   = value[key],
                        reason  = self.CODE_EXTRA_KEY,
                        sub_key = unicode_key,

                        # https://github.com/eflglobal/filters/issues/15
                        template_vars = {
                            'actual_key': unicode_key,
                        },
                    )

    def _apply_item(self, key, value, filter_chain):
        # type: (Text, Any, FilterCompatible) -> Any
        """
        Applies filters to a single item in the mapping.

        Override this method in a subclass if you want to customize the
        way specific items get filtered.
        """
        return self._filter(value, filter_chain, sub_key=key)

    def _missing_key_allowed(self, key):
        # type: (Text) -> bool
        """
        Returns whether the specified key is allowed to be omitted from
        the incoming value.
        """
        if self.allow_missing_keys is True:
            return True

        try:
            return key in self.allow_missing_keys
        except TypeError:
            return False

    def _extra_key_allowed(self, key):
        # type: (Text) -> bool
        """
        Returns whether the specified extra key is allowed.
        """
        if self.allow_extra_keys is True:
            return True

        try:
            return key in self.allow_extra_keys
        except TypeError:
            return False

    @staticmethod
    def unicodify_key(key):
        # type: (Text) -> Text
        """
        Converts a key value into a unicode so that it can be
        represented in e.g., error message contexts.
        """
        if key is None:
            return 'None'

        try:
            return Unicode().apply(key)
        except FilterError:
            return repr(key)



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import sys
from collections import OrderedDict
from logging import ERROR, Logger, LoggerAdapter
from traceback import format_exc
from types import TracebackType
from typing import Any, Dict, List, Text, Tuple, Union

from six import iteritems, python_2_unicode_compatible, text_type

from filters.base import BaseFilter, BaseInvalidValueHandler, FilterCompatible

__all__ = [
    'FilterMessage',
    'FilterRunner',
    'LogHandler',
    'MemoryHandler',
]


class LogHandler(BaseInvalidValueHandler):
    """
    Invalid value handler that sends the details to a logger.
    """
    def __init__(self, logger, level=ERROR):
        # type: (Union[Logger, LoggerAdapter], int) -> None
        """
        :param logger: The logger that log messages will get sent to.
        :param level: Level of the logged messages.
        """
        super(LogHandler, self).__init__()

        self.logger = logger
        self.level  = level

    def handle_invalid_value(self, message, exc_info, context):
        self.logger.log(
            level       = self.level,
            msg         = message,
            exc_info    = exc_info,
            extra       = {'context': context}
        )


@python_2_unicode_compatible
class FilterMessage(object):
    """Provides a consistent API for messages sent to MemoryHandler."""
    def __init__(self, message, context, exc_info=None):
        # type: (Text, dict, Text) -> None
        """
        :param exc_info: Exception traceback (if applicable).
        """
        super(FilterMessage, self).__init__()

        self.message    = message
        self.context    = context
        self.code       = context.get('code') or message
        self.exc_info   = exc_info

    def __repr__(self):
        return '{type}({message}, {context})'.format(
            type    = type(self).__name__,
            message = repr(self.message),
            context = repr(self.context),
        )

    def __str__(self):
        return self.message

    def as_dict(self, with_debug_info=False):
        # type: (bool) -> dict
        """
        Returns a dict representation of the FilterMessage.

        :param with_debug_info:
            Whether to include context and exception traceback in the
            result.
        """
        res = {
            'code':     self.code,
            'message':  self.message,
        }

        if with_debug_info:
            res['context']  = self.context
            res['exc_info'] = self.exc_info

        return res


class MemoryHandler(BaseInvalidValueHandler):
    """Invalid value handler that stores messages locally."""
    def __init__(self, capture_exc_info=False):
        # type: (bool) -> None
        """
        :param capture_exc_info:
            Whether to capture `sys.exc_info` when an handling an
            exception.

            This is turned off by default to reduce memory usage, but
            it is useful in certain cases (e.g., if you want to send
            exceptions to a logger that expect exc_info).

            Regardless, you can still check ``self.has_exceptions`` to
            see if an exception occurred.
        """
        super(MemoryHandler, self).__init__()

        self.messages           = OrderedDict() # type: Union[OrderedDict, Dict[Text, List[FilterMessage]]]
        self.has_exceptions     = False
        self.capture_exc_info   = capture_exc_info
        self.exc_info           = [] # type: List[Tuple[type, Exception, TracebackType]]

    def handle_invalid_value(self, message, exc_info, context):
        key = context.get('key', '')
        msg = FilterMessage(
            message     = message,
            context     = context,
            exc_info    = format_exc() if exc_info else None,
        )

        try:
            self.messages[key].append(msg)
        except KeyError:
            self.messages[key] = [msg]

    def handle_exception(self, message, exc):
        self.has_exceptions = True

        if self.capture_exc_info:
            self.exc_info.append(sys.exc_info())

        return super(MemoryHandler, self).handle_exception(message, exc)


@python_2_unicode_compatible
class FilterRunner(object):
    """
    Wrapper for a filter that provides an API similar to what you would
    expect from a Django form (at least, when it comes to methods
    related to data validation).

    Note that FilterRunner is intended to be a "one-shot" tool; once
    initialized, it does not expect the data it is filtering to
    change.
    """
    def __init__(self, starting_filter, incoming_data, capture_exc_info=False):
        # type: (FilterCompatible, Any, bool) -> None
        """
        :param incoming_data: E.g., `request.POST`.

        :param capture_exc_info:
            Whether to capture `sys.exc_info` when an handling an
            exception.

            This is turned off by default to reduce memory usage, but
            it is useful in certain cases (e.g., if you want to send
            exceptions to a logger).

            Regardless, you can still check ``self.has_exceptions`` to
            see if an exception occurred.
        """
        super(FilterRunner, self).__init__()

        self.filter_chain       = BaseFilter.resolve_filter(starting_filter) # type: FilterCompatible
        self.data               = incoming_data
        self.capture_exc_info   = capture_exc_info

        self._cleaned_data  = None
        self._handler       = None # type: MemoryHandler


    def __str__(self):
        return text_type(self.filter_chain)

    @property
    def cleaned_data(self):
        """
        Returns the resulting value after applying the request filter.
        """
        self.full_clean()
        return self._cleaned_data

    @property
    def errors(self):
        # type: () -> Dict[Text, List[Dict[Text, Text]]]
        """
        Returns a dict of error messages generated by the Filter, in a
        format suitable for inclusion in e.g., an API 400 response
        payload.

        E.g.::

            {
                'authToken': [
                    {
                        'code':     'not_found',
                        'message':
                            'No AuthToken found matching this value.',
                    },
                ],

                'data.foobar': [
                    {
                        'code':     'unexpected',
                        'message':  'Unexpected key "foobar".',
                    },
                ],

                # etc.
            }
        """
        return self.get_errors()

    def get_errors(self, with_context=False):
        # type: (bool) -> Dict[Text, List[Dict[Text, Text]]]
        """
        Returns a dict of error messages generated by the Filter, in a
        format suitable for inclusion in e.g., an API 400 response
        payload.

        :param with_context:
            Whether to include the context object in the result (for
            debugging purposes).

            Note: context is usually not safe to expose to end users!
        """
        return {
            key: [m.as_dict(with_context) for m in messages]
                for key, messages in iteritems(self.filter_messages)
        }

    @property
    def error_codes(self):
        # type: () -> Dict[Text, List[Text]]
        """
        Returns a dict of error codes generated by the Filter.
        """
        return {
            key: [m.code for m in messages]
                for key, messages in iteritems(self.filter_messages)
        }

    @property
    def has_exceptions(self):
        # type: () -> bool
        """
        Returns whether any unhandled exceptions occurred while
        filtering the value.
        """
        self.full_clean()
        return self._handler.has_exceptions

    @property
    def exc_info(self):
        # type: () -> List[Tuple[type, Exception, TracebackType]]
        """
        Returns tracebacks from any exceptions that were captured.
        """
        self.full_clean()
        return self._handler.exc_info

    @property
    def filter_messages(self):
        # type: () -> Dict[Text, List[FilterMessage]]
        """
        Returns the raw FilterMessages that were generated by the
        Filter.
        """
        self.full_clean()
        return self._handler.messages

    def is_valid(self):
        # type: () -> bool
        """
        Returns whether the request payload successfully passed the
        filter.
        """
        return not self.filter_messages

    def full_clean(self):
        """Applies the filter to the request data."""
        if self._handler is None:
            self._handler = MemoryHandler(self.capture_exc_info)

            # Inject our own handler (temporarily) while the Filter
            # runs so we can capture error messages.
            prev_handler = self.filter_chain.handler
            self.filter_chain.handler = self._handler
            try:
                self._cleaned_data = self.filter_chain.apply(self.data)
            finally:
                self.filter_chain.handler = prev_handler



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import json
import re
import socket
import unicodedata
from base64 import standard_b64decode, urlsafe_b64decode
from collections import OrderedDict
from decimal import Decimal as DecimalType
from typing import Any, Callable, Optional, Sequence, Text, Union
from uuid import UUID
from xml.etree.ElementTree import Element, tostring

# noinspection PyCompatibility
import regex
from six import PY2, PY3, binary_type, moves as compat, \
    python_2_unicode_compatible, text_type

from filters.base import BaseFilter, Type
from filters.simple import MaxLength

__all__ = [
    'Base64Decode',
    'ByteString',
    'CaseFold',
    'IpAddress',
    'JsonDecode',
    'MaxBytes',
    'Regex',
    'Split',
    'Strip',
    'Unicode',
    'Uuid',
]


class Base64Decode(BaseFilter):
    """
    Decodes an incoming value using the Base64 algo.
    """
    CODE_INVALID = 'not_base64'

    templates = {
        CODE_INVALID: 'Base64-encoded value expected.',
    }

    def __init__(self):
        super(Base64Decode, self).__init__()

        self.whitespace_re = regex.compile(b'[ \t\r\n]+', regex.ASCII)
        self.base64_re = regex.compile(b'^[-+_/A-Za-z0-9=]+$', regex.ASCII)

    def _apply(self, value):
        value = self._filter(value, Type(binary_type)) # type: binary_type

        if self._has_errors:
            return None

        # Strip out whitespace.
        # Technically, whitespace is not part of the Base64 alphabet,
        # but virtually every implementation allows it.
        value = self.whitespace_re.sub(b'', value)

        # Check for invalid characters.
        # Note that Python 3's b64decode does this for us, but we also
        # have to support Python 2.
        # https://docs.python.org/3/library/base64.html#base64.b64decode
        if not self.base64_re.match(value):
            return self._invalid_value(
                value   = value,
                reason  = self.CODE_INVALID,
            )

        # Check to see if we are working with a URL-safe dialect.
        # https://en.wikipedia.org/wiki/Base64#URL_applications
        if (b'_' in value) or (b'-' in value):
            # You can't mix dialects, silly!
            if (b'+' in value) or (b'/' in value):
                return self._invalid_value(
                    value   = value,
                    reason  = self.CODE_INVALID,
                )

            url_safe = True
        else:
            url_safe = False

        # Normalize padding.
        # http://stackoverflow.com/a/9807138/
        value = value.rstrip(b'=')
        value += (b'=' * (4 - (len(value) % 4)))

        try:
            return (
                urlsafe_b64decode(value)
                    if url_safe
                    else standard_b64decode(value)
            )
        except TypeError:
            return self._invalid_value(value, self.CODE_INVALID, exc_info=True)


# noinspection SpellCheckingInspection
class CaseFold(BaseFilter):
    """
    Applies case folding to an incoming string, allowing you to perform
    case-insensitive comparisons.

    The result tends to be lowercase, but it is recommended that you
    NOT treat CaseFold as a Unicode-aware lowercase filter!  The
    proper way to lowercase a string is very much locale-dependent.

    Note that the built in :py:meth:`str.upper` and
    :py:meth:`str.lower` methods tend do a pretty good job of properly
    changing the case of unicode strings.

    References:
      - http://www.w3.org/International/wiki/Case_folding
      - https://docs.python.org/3/library/stdtypes.html#str.lower
      - https://docs.python.org/3/library/stdtypes.html#str.upper
    """
    def _apply(self, value):
        value = self._filter(value, Type(text_type)) # type: Text

        if self._has_errors:
            return None

        # In Python 3, case folding is supported natively.
        # In Python 2, this is the best we can do.
        # https://docs.python.org/3/library/stdtypes.html#str.casefold
        if PY3:
            # noinspection PyUnresolvedReferences
            return value.casefold()
        else:
            # noinspection PyUnresolvedReferences
            from py2casefold import casefold
            return casefold(value)


@python_2_unicode_compatible
class IpAddress(BaseFilter):
    """
    Validates an incoming value as an IPv[46] address.
    """
    CODE_INVALID = 'not_ip_address'

    templates = {
        CODE_INVALID: 'This value is not a valid {ip_type} address.',
    }

    def __init__(self, ipv4=True, ipv6=False):
        # type: (bool, bool) -> None
        super(IpAddress, self).__init__()

        self.ipv4 = ipv4
        self.ipv6 = ipv6

    def __str__(self):
        return '{type}(ipv4={ipv4!r}, ipv6={ipv6!r})'.format(
            type    = type(self).__name__,
            ipv4    = self.ipv4,
            ipv6    = self.ipv6,
        )

    @property
    def ip_type(self):
        # type: () -> Text
        """
        Returns the IP address versions that this Filter accepts.
        """
        return '/'.join(filter(None, [
            'IPv4' if self.ipv4 else None,
            'IPv6' if self.ipv6 else None,
        ]))

    def _apply(self, value):
        value = self._filter(value, Type(text_type))

        if self._has_errors:
            return None

        # http://stackoverflow.com/a/4017219
        if self.ipv4:
            try:
                socket.inet_pton(socket.AF_INET, value)
            except socket.error:
                pass
            else:
                return value

        if self.ipv6:
            try:
                n = socket.inet_pton(socket.AF_INET6, value)
            except socket.error:
                pass
            else:
                # Convert the binary value back into a string
                # representation so that the end result is
                # normalized.
                # https://en.wikipedia.org/wiki/IPv6_address#Presentation
                return socket.inet_ntop(socket.AF_INET6, n)

        # If we get here, we failed the above checks (or the Filter is
        # configured not to allow anything through).
        return self._invalid_value(
            value   = value,
            reason  = self.CODE_INVALID,

            template_vars = {
                'ip_type':  self.ip_type
            },
        )


class JsonDecode(BaseFilter):
    """
    Interprets the value as JSON.

    JSON objects are converted to OrderedDict instances so that key
    order is preserved.
    """
    CODE_INVALID = 'not_json'

    templates = {
        CODE_INVALID: 'This value is not valid JSON.',
    }

    def __init__(self, decoder=json.loads):
        # type: (Callable[Text, Any]) -> None
        super(JsonDecode, self).__init__()

        self.decoder = decoder

    def _apply(self, value):
        value = self._filter(value, Type(text_type)) # type: Text

        if self._has_errors:
            return None

        try:
            # :see: http://stackoverflow.com/a/6921760
            return self.decoder(value, object_pairs_hook=OrderedDict)
        except ValueError:
            return self._invalid_value(value, self.CODE_INVALID, exc_info=True)


@python_2_unicode_compatible
class MaxBytes(BaseFilter):
    """
    Ensures that an incoming string value is small enough to fit into a
    specified number of bytes when encoded.

    Note:  The resulting value is a byte string, even if you provide a
    unicode.
    """
    CODE_TOO_LONG = 'too_long'

    templates = {
        CODE_TOO_LONG:
            'Value is too long (must be < {max_bytes} '
            'bytes when encoded using {encoding}).',
    }

    def __init__(
            self,
            max_bytes,
            truncate    = True,
            prefix      = '',
            encoding    = 'utf-8',
    ):
        # type: (int, bool, Text, Text) -> None
        """
        :param max_bytes:
            Max number of bytes to allow.

        :param truncate:
            Whether to truncate values that are too long.

            Set this to ``False`` to save system resources when you
            know that you will reject values that are too long.

        :param prefix:
            Prefix to apply to truncated values.
            
            Ignored when ``truncate`` is ``False``.

        :param encoding:
            The character encoding to check against.

            Note: This filter is optimized for UTF-8.
        """
        super(MaxBytes, self).__init__()

        self.encoding   = encoding
        self.max_bytes  = max_bytes
        self.prefix     = prefix
        self.truncate   = truncate

    def __str__(self):
        return '{type}({max_bytes!r}, encoding={encoding!r})'.format(
            type        = type(self).__name__,
            max_bytes   = self.max_bytes,
            encoding    = self.encoding,
        )

    def _apply(self, value):
        """
        :return:
            Returns bytes, truncated to the correct length.

            Note: Might be a bit shorter than the max length, to avoid
            orphaning a multibyte sequence.
        """
        value = self._filter(
            value = value,

            filter_chain = (
                    Type((binary_type, text_type,))
                |   Unicode(encoding=self.encoding)
            ),
        ) # type: Text

        if self._has_errors:
            return None

        str_value = value.encode(self.encoding)

        if len(str_value) > self.max_bytes:
            replacement = (
                self.truncate_string(
                    # Ensure that we convert back to unicode before
                    # adding the prefix, just in case `self.encoding`
                    # indicates a codec that uses a BOM.
                    value = self.prefix + value,

                    max_bytes   = self.max_bytes,
                    encoding    = self.encoding,
                )
                    if self.truncate
                    else None
            )

            return self._invalid_value(
                value       = value,
                reason      = self.CODE_TOO_LONG,
                replacement = replacement,

                context = {
                    'encoding':     self.encoding,
                    'max_bytes':    self.max_bytes,
                    'prefix':       self.prefix,
                    'truncate':     self.truncate,
                },
            )

        return str_value

    @staticmethod
    def truncate_string(value, max_bytes, encoding):
        # type: (Text, int, Text) -> binary_type
        """
        Truncates a string value to the specified number of bytes.

        :return:
            Returns bytes, truncated to the correct length.

            Note: Might be a bit shorter than `max_bytes`, to avoid
            orphaning a multibyte sequence.
        """
        # Convert to bytearray so that we get the same handling in
        # Python 2 and Python 3.
        bytes_ = bytearray(value.encode(encoding))

        # Truncating the value is a bit tricky, as we have to be
        # careful not to leave an unterminated multibyte sequence.

        if encoding.lower() in ['utf-8', 'utf8']:
            #
            # This code works a bit faster than the generic routine
            # (see below) because we only have to inspect up to 4
            # bytes from the end of the encoded value instead of
            # having to repeatedly decode the entire string.
            #
            # But, it only works for UTF-8.
            #
            truncated = bytes_[0:max_bytes]

            # Walk backwards through the string until we hit certain
            # sequences.
            for i, o in enumerate(reversed(truncated), start=1):
                # If the final byte is not part of a multibyte
                # sequence, then we can stop right away; there is no
                # need to remove anything.
                if (i < 2) and (o < 0b10000000):
                    break

                # If this byte is a leading byte (the first byte in a
                # multibyte sequence), determine how many bytes we
                # need to strip off the end of the string so that we
                # can decode it back into a unicode if needed.
                if o >= 0b11000000:
                    # Note:  Assuming max 4 bytes per sequence.
                    # Should be good enough until extraterrestrial
                    # languages are encountered.
                    seq_length = (
                        4 if o >= 0b11110000 else
                        3 if o >= 0b11100000 else
                        2
                    )

                    # Now that we know how many bytes are in the final
                    # sequence, check to see if it is complete, and
                    # discard it if it is incomplete.
                    if seq_length != i:
                        truncated = truncated[0:-i]

                    break

                # Else, we have a continuation byte.  Continue walking
                # backwards through the string.

            return truncated

        else:
            trim = 0
            while True:
                # Progressively chop bytes off the end of the string
                # until we have something that can be successfully
                # decoded using the specified encoding.
                truncated = bytes_[0:max_bytes - trim]

                try:
                    truncated.decode(encoding)
                except UnicodeDecodeError:
                    trim += 1
                else:
                    return binary_type(truncated)

                # We should never get here, but just in case, we need
                # to ensure the loop eventually terminates (Python
                # won't error if ``max_bytes - trim`` goes negative,
                # since the slice operator accepts negative values).
                if trim >= max_bytes:
                    raise ValueError(
                        'Unable to truncate {bytes_!r} to {max_bytes} '
                        'bytes when encoded using {encoding}.'.format(
                            bytes_      = bytes_,
                            max_bytes   = max_bytes,
                            encoding    = encoding,
                        ),
                    )


@python_2_unicode_compatible
class Regex(BaseFilter):
    """
    Matches a regular expression in the value.

    IMPORTANT: This filter returns a LIST of all sequences in the
    input value that matched the regex!

    IMPORTANT: This Filter uses the ``regex`` library, which behaves
    slightly differently than Python's ``re`` library.

    If you've never used ``regex`` before, try it; you'll never want to
    go back!

    References:
      - https://pypi.python.org/pypi/regex
    """
    CODE_INVALID = 'malformed'

    templates = {
        CODE_INVALID:
            'Value does not match regular expression {pattern}.',
    }

    # noinspection PyProtectedMember
    def __init__(self, pattern):
        # type: (Union[Text, regex._pattern_type, re._pattern_type]) -> None
        """
        :param pattern:
            String pattern, or pre-compiled regex.

            IMPORTANT:  If you specify your own compiled regex, be sure to
            add the ``UNICODE`` flag for Unicode support!
        """
        super(Regex, self).__init__()

        self.regex = (
            pattern
                if isinstance(pattern, (regex._pattern_type, re._pattern_type))
                else regex.compile(pattern, regex.UNICODE)
        )

    def __str__(self):
        return '{type}({pattern!r})'.format(
            type    = type(self).__name__,
            pattern = self.regex.pattern,
        )

    def _apply(self, value):
        value = self._filter(value, Type(text_type))

        if self._has_errors:
            return None

        matches = [
            match.group(0)
                for match in self.regex.finditer(value)
        ]

        if not matches:
            return self._invalid_value(
                value   = value,
                reason  = self.CODE_INVALID,

                template_vars = {
                    'pattern': self.regex.pattern,
                },
            )

        return matches


@python_2_unicode_compatible
class Split(BaseFilter):
    """
    Splits an incoming string into parts.

    The result is either a list or an OrderedDict, depending on whether
    you specify keys to map to the result.
    """
    # noinspection PyProtectedMember
    def __init__(self, pattern, keys=None):
        # type: (Union[Text, regex._pattern_type, re._pattern_type], Optional[Sequence[Text]]) -> None
        """
        :param pattern:
            Regex used to split incoming string values.

            IMPORTANT:  If you specify your own compiled regex, be sure
            to add the ``UNICODE`` flag for Unicode support!

        :param keys:
            If set, the resulting list will be converted into an
            OrderedDict, using the specified keys.

            IMPORTANT:  If ``keys`` is set, the split value's length
            must be less than or equal to ``len(keys)``.
        """
        super(Split, self).__init__()

        self.regex = (
            pattern
                if isinstance(pattern, (regex._pattern_type, re._pattern_type))
                else regex.compile(pattern, regex.UNICODE)
        )

        self.keys = keys

    def __str__(self):
        return '{type}({pattern!r}, keys={keys!r}'.format(
            type    = type(self).__name__,
            pattern = self.regex.pattern,
            keys    = self.keys,
        )

    def _apply(self, value):
        value = self._filter(value, Type(text_type))

        if self._has_errors:
            return None

        split = self.regex.split(value)

        if self.keys:
            # The split value can have at most as many items as
            # ``self.keys``.
            split = self._filter(split, MaxLength(len(self.keys)))

            if self._has_errors:
                return None

            return OrderedDict(compat.zip_longest(self.keys, split))
        else:
            return split


@python_2_unicode_compatible
class Strip(BaseFilter):
    """
    Strips characters (whitespace and non-printables by default) from
    the end(s) of a string.

    IMPORTANT:  This Filter uses the ``regex`` library, which behaves
    slightly differently than Python's ``re`` library.

    If you've never used ``regex`` before, try it; you'll never want to
    go back!
    """
    def __init__(self, leading=r'[\p{C}\s]+', trailing=r'[\p{C}\s]+'):
        # type: (Text, Text) -> None
        """
        :param leading:
            Regex to match at the start of the string.
            
        :param trailing:
            Regex to match at the end of the string.
        """
        super(Strip, self).__init__()

        if leading:
            self.leading = regex.compile(
                r'^{pattern}'.format(pattern=leading),
                regex.UNICODE,
            )
        else:
            self.leading = None

        if trailing:
            self.trailing = regex.compile(
                r'{pattern}$'.format(pattern=trailing),
                regex.UNICODE,
            )
        else:
            self.trailing = None

    def __str__(self):
        return '{type}(leading={leading!r}, trailing={trailing!r})'.format(
            type        = type(self).__name__,
            leading     = self.leading.pattern,
            trailing    = self.trailing.pattern,
        )

    def _apply(self, value):
        value = self._filter(value, Type(text_type))

        if self._has_errors:
            return None

        if self.leading:
            value = self.leading.sub('', value)

        if self.trailing:
            value = self.trailing.sub('', value)

        return value


@python_2_unicode_compatible
class Unicode(BaseFilter):
    """
    Converts a value into a unicode string.

    Note:  By default, additional normalization is applied to the
    resulting value.  See the initializer docstring for more info.

    References:
      - https://docs.python.org/2/howto/unicode.html
      - https://en.wikipedia.org/wiki/Unicode_equivalence
    """
    CODE_DECODE_ERROR = 'wrong_encoding'

    templates = {
        CODE_DECODE_ERROR: 'This value cannot be decoded using {encoding}.',
    }

    def __init__(self, encoding='utf-8', normalize=True):
        # type: (Text, bool) -> None
        """
        :param encoding:
            Used to decode non-unicode values.

        :param normalize:
            Whether to normalize the resulting value:
                - Convert to NFC form.
                - Remove non-printable characters.
                - Convert all line endings to unix-style ('\n').
        """
        super(Unicode, self).__init__()

        self.encoding   = encoding
        self.normalize  = normalize

        if self.normalize:
            #
            # Compile the regex that we will use to remove non-
            # printables from the resulting unicode.
            # http://www.regular-expressions.info/unicode.html#category
            #
            # Note: using a double negative so that we can exclude
            # newlines, which are technically considered control chars.
            # http://stackoverflow.com/a/3469155
            #
            self.npr = regex.compile(r'[^\P{C}\s]+', regex.UNICODE)

    def __str__(self):
        return '{type}(encoding={encoding!r})'.format(
            type        = type(self).__name__,
            encoding    = self.encoding,
        )

    def _apply(self, value):
        try:
            if isinstance(value, text_type):
                decoded = value

            elif isinstance(value, binary_type):
                decoded = value.decode(self.encoding)

            elif isinstance(value, bool):
                decoded = text_type(int(value))

            # In Python 3, ``bytes(<int>)`` does weird things.
            # https://www.python.org/dev/peps/pep-0467/
            elif isinstance(value, (int, float)):
                decoded = text_type(value)

            elif isinstance(value, DecimalType):
                decoded = format(value, 'f')

            elif isinstance(value, Element):
                # There's no way (that I know of) to get
                # :py:meth:`ElementTree.tostring` to return a unicode.
                decoded = tostring(value, 'utf-8').decode('utf-8')

            elif (
                    PY2 and hasattr(value, '__str__')
                or  PY3 and hasattr(value, '__bytes__')
            ):
                decoded = binary_type(value).decode(self.encoding)

            else:
                decoded = text_type(value)
        except UnicodeDecodeError:
            return self._invalid_value(
                value           = value,
                reason          = self.CODE_DECODE_ERROR,
                exc_info        = True,

                template_vars = {
                    'encoding': self.encoding,
                },
            )

        if self.normalize:
            return (
                # Return the final string in composed form.
                # https://en.wikipedia.org/wiki/Unicode_equivalence
                unicodedata.normalize('NFC',
                    # Remove non-printables.
                    self.npr.sub('', decoded)
                )
                    # Normalize line endings.
                    # http://stackoverflow.com/a/1749887
                    .replace('\r\n', '\n')
                    .replace('\r', '\n')
            )
        else:
            return decoded


class ByteString(Unicode):
    """
    Converts a value into a byte string, encoded as UTF-8.

    IMPORTANT: This filter returns bytes objects, not bytearrays!
    """
    def __init__(self, encoding='utf-8', normalize=False):
        # type: (Text, bool) -> None
        """
        :param encoding:
            Used to decode non-unicode values.

        :param normalize:
            Whether to normalize the unicode value before converting
            back into bytes:

                - Convert to NFC form.
                - Remove non-printable characters.
                - Convert all line endings to unix-style ('\n').

            Note that ``normalize`` is ``False`` by default for
            :py:class:`ByteString`, but ``True`` by default for
            :py:class:`Unicode`.
        """
        super(ByteString, self).__init__(encoding, normalize)

    # noinspection SpellCheckingInspection
    def _apply(self, value):
        decoded = super(ByteString, self)._apply(value) # type: Text

        #
        # No need to catch UnicodeEncodeErrors here; UTF-8 can handle
        # any unicode value.
        #
        # Technically, we could get this error if we encounter a code
        # point beyond U+10FFFF (the highest valid code point in the
        # Unicode standard).
        #
        # However, it's not possible to create a `unicode` object with
        # an invalid code point, so we wouldn't even be able to get
        # this far if the incoming value contained a character that
        # can't be represented using UTF-8.
        #
        # Note that in some versions of Python, it is possible (albeit
        # really difficult) to trick Python into creating unicode
        # objects with invalid code points, but it generally requires
        # using specific codecs that aren't UTF-8.
        #
        # Example of exploit and release notes from the Python release
        # (2.7.6) that fixes the issue:
        #
        #   - https://gist.github.com/rspeer/7559750
        #   - https://hg.python.org/cpython/raw-file/99d03261c1ba/Misc/NEWS
        #

        # Normally we return ``None`` if we get any errors, but in this
        # case, we'll let the superclass method decide.
        return decoded if self._has_errors else decoded.encode('utf-8')


@python_2_unicode_compatible
class Uuid(BaseFilter):
    """
    Interprets an incoming value as a UUID.
    """
    CODE_INVALID        = 'not_uuid'
    CODE_WRONG_VERSION  = 'wrong_version'

    templates = {
        CODE_INVALID: 'This value is not a well-formed UUID.',

        CODE_WRONG_VERSION:
            'v{incoming} UUID not allowed (expected v{expected}).',
    }

    def __init__(self, version=None):
        # type: (Optional[int]) -> None
        """
        :type version:
            If specified, requires the resulting UUID to match the
            specified version.

        References:
          - https://en.wikipedia.org/wiki/Uuid#RFC_4122_Variant
        """
        super(Uuid, self).__init__()

        self.version = version

    def __str__(self):
        return '{type}(version={version!r})'.format(
            type    = type(self).__name__,
            version = self.version,
        )

    def _apply(self, value):
        value = self._filter(value, Type((text_type, UUID,))) # type: Union[Text, UUID]

        if self._has_errors:
            return None

        try:
            uuid = (
                value
                    if isinstance(value, UUID)
                    else UUID(hex=value)
            )
        except ValueError:
            return self._invalid_value(value, self.CODE_INVALID, exc_info=True)
        else:
            if self.version not in (None, uuid.version):
                return self._invalid_value(
                    value   = text_type(uuid),
                    reason  = self.CODE_WRONG_VERSION,

                    context = {
                        'expected': self.version,
                        'incoming': uuid.version,
                    },
                )

            return uuid



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from inspect import getmembers as get_members, isabstract as is_abstract, \
    isclass as is_class, ismodule as is_module
from logging import getLogger
from typing import Any, Dict, Generator, Optional, Text, Tuple, Type, Union

from pkg_resources import EntryPoint, iter_entry_points
from six import iterkeys, python_2_unicode_compatible, text_type

from filters.base import BaseFilter

ENTRY_POINT_KEY = 'filters.extensions'
"""
The key to use when declaring entry points in your library's
``setup.py`` file.

Example::

   setup(
     ...
     entry_points = {
       'filters.extensions': [
         # Load all filters from a single module.
         'iso = filters_iso',

         # Load a single class.
         'currency = filters_iso:Currency',
       ],
     },
   )

Filters that are loaded this way are accessible from
:py:data:`filters.ext` (not imported into the global namespace because
it gives IDEs a heart attack).
"""

logger = getLogger(__name__)

@python_2_unicode_compatible
class FilterExtensionRegistry(object):
    """
    Creates a registry that can be used to dynamically load 3rd-party
    filters into the (nearly) top-level namespace.
    """
    def __init__(self, filters=None):
        # type: (Optional[Dict[Text, Type[BaseFilter]]]) -> None
        """
        :param filters:
            Used to preload the registry.

            If ``None``, filters will be automatically loaded from
            3rd-party libraries.

            If you want to initialize a completely empty registry,
            set ``filters`` to an empty dict.
        """
        super(FilterExtensionRegistry, self).__init__()

        self._filters = filters

    def __dir__(self):
        self.__autoload()
        return list(iterkeys(self._filters))

    def __getattr__(self, item):
        return self[item]

    def __getitem__(self, item):
        self.__autoload()
        return self._filters.get(item)

    def __iter__(self):
        self.__autoload()
        return iter(self._filters)

    def __missing__(self, key):
        raise KeyError('Extension filter "{key}" not found!'.format(key=key))

    def __repr__(self):
        self.__autoload()
        return repr(self._filters)

    def __str__(self):
        self.__autoload()
        return text_type(self._filters)

    def __autoload(self):
        """
        Automatically loads registered extension filters, if necessary.
        """
        if self._filters is None:
            self._filters = discover_filters()


def discover_filters():
    # type: () -> Dict[Text, Type[BaseFilter]]
    """
    Returns all registered filters, in no particular order.

    If two filters have the same name, the one that gets loaded second
    will replace the first one.  Note that the order that filters are
    loaded is not defined.
    """
    filters = {} # type: Dict[Text, Type[BaseFilter]]

    for entry_point in iter_entry_points(ENTRY_POINT_KEY): # type: EntryPoint
        logger.debug(
            'Looking for extension filters in {name} (`{target}`).'.format(
                name    = entry_point.name,
                target  = entry_point.module_name,
            ),
        )

        filters.update(iter_filters_in(entry_point.resolve()))

    return filters


def is_filter_type(target):
    # type: (Any) -> Union[bool, Text]
    """
    Returns whether the specified object can be registered as a filter.

    :return:
        Returns ``True`` if the object is a filter.
        Otherwise, returns a string indicating why it is not valid.
    """
    if not is_class(target):
        return 'not a class'

    if not issubclass(target, BaseFilter):
        return 'does not extend BaseFilter'

    if is_abstract(target):
        return 'abstract class'

    return True


def iter_filters_in(target):
    # type: (Any) -> Generator[Tuple[Text, Type[BaseFilter]]]
    """
    Iterates over all filters in the specified module/class.
    """
    ift_result = is_filter_type(target)

    if ift_result is True:
        logger.debug(
            'Registering extension filter '
            '{cls.__module__}.{cls.__name__}.'.format(
                cls = target,
            ),
        )

        yield target.__name__, target
    elif is_module(target):
        for member_name, member in get_members(target):
            member_ift_result = is_filter_type(member)

            if member_ift_result is True:
                logger.debug(
                    'Registering extension filter '
                    '{cls.__module__}.{cls.__name__}.'.format(
                        cls = member,
                    ),
                )

                yield member.__name__, member
            else:
                logger.debug(
                    'Ignoring {module}.{name} ({reason})'.format(
                        module  = target.__name__,
                        name    = member_name,
                        reason  = member_ift_result,
                    ),
                )
    elif is_class(target):
        logger.debug(
            'Ignoring {cls.__module__}.{cls.__name__} ({reason}).'.format(
                cls     = target,
                reason  = ift_result,
            ),
        )
    else:
        logger.debug(
            'Ignoring {target!r} ({reason}).'.format(
                reason  = ift_result,
                target  = target,
            ),
        )



# coding=utf-8
# :bc: Not importing unicode_literals because in Python 2 distutils,
# some values are expected to be byte strings.
from __future__ import absolute_import, division, print_function

from codecs import StreamReader, open
from os.path import dirname, join, realpath
from sys import version_info

from setuptools import setup

cwd = dirname(realpath(__file__))

##
# Load long description for PyPi.
with open(join(cwd, 'README.rst'), 'r', 'utf-8') as f: # type: StreamReader
    long_description = f.read()


##
# For compatibility with versions of pip < 9, we will determine
# dependencies at runtime.
# Maybe once Travis upgrades their containers to use a newer version,
# we'll switch to the newer syntax (:
dependencies = [
    'python-dateutil',
    'pytz',
    'regex',
    'six',

]

if version_info[0] < 3:
    # noinspection SpellCheckingInspection
    dependencies.extend([
        'py2casefold', # 'py2casefold; python_version < "3.0"',
        'typing', # 'typing; python_version < "3.0"',
    ])


##
# Off we go!
setup(
    name        = 'filters',
    description = 'Validation and data pipelines made easy!',
    url         = 'https://filters.readthedocs.io/',

    version = '1.2.3',

    packages = ['filters'],

    long_description = long_description,

    install_requires = dependencies,

    extras_require = {
        'django':['filters-django'],
        'iso': ['filters-iso'],
    },

    test_suite    = 'test',
    test_loader   = 'nose.loader:TestLoader',
    tests_require = [
        'nose',
    ],

    data_files = [
        ('', ['LICENSE']),
    ],

    license = 'MIT',

    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Filters',
    ],

    keywords = 'data validation',

    author          = 'Phoenix Zerin',
    author_email    = 'phoenix.zerin@eflglobal.com',
)



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from decimal import Decimal as DecimalType, InvalidOperation, ROUND_HALF_UP

from six import python_2_unicode_compatible, text_type
from typing import Any, Text, Union

from filters.base import BaseFilter, Type

__all__ = [
    'Decimal',
    'Int',
    'Max',
    'Min',
    'Round',
]


@python_2_unicode_compatible
class Decimal(BaseFilter):
    """
    Interprets the value as a :py:class:`decimal.Decimal` object.
    """
    CODE_INVALID    = 'not_numeric'
    CODE_NON_FINITE = 'not_finite'

    templates = {
        CODE_INVALID:       'Numeric value expected.',
        CODE_NON_FINITE:    'Numeric value expected.',
    }

    def __init__(self, max_precision=None, allow_tuples=True):
        # type: (Union[int, DecimalType], bool) -> None
        """
        :param max_precision:
            Max number of decimal places the resulting value is allowed
            to have.  Values that are too precise will be rounded to
            fit.

            To avoid ambiguity, specify ``max_precision`` as a
            ``decimal.Decimal`` object.

            For example, to round to the nearest 1/100::

                Decimal(max_precision=decimal.Decimal('0.01'))

        :param allow_tuples:
            Whether to allow tuple-like inputs.

            Allowing tuple inputs might couple the implementation more
            tightly to Python's Decimal type, so you have the option
            to disallow it.
        """
        super(Decimal, self).__init__()

        # Convert e.g., 3 => DecimalType('.001').
        if not (
                    (max_precision is None)
                or  isinstance(max_precision, DecimalType)
        ):
            max_precision = DecimalType('.1') ** max_precision

        self.max_precision  = max_precision
        self.allow_tuples   = allow_tuples

    def __str__(self):
        return '{type}(max_precision={max_precision!r})'.format(
            type            = type(self).__name__,
            max_precision   = self.max_precision,
        )

    def _apply(self, value):
        allowed_types = (text_type, int, float, DecimalType,)
        if self.allow_tuples:
            # Python's Decimal type supports both tuples and lists.
            # :py:meth:`decimal.Decimal.__init__`
            allowed_types += (list, tuple,)

        value = self._filter(value, Type(allowed_types))

        if self._has_errors:
            return value

        try:
            d = DecimalType(value)
        except (InvalidOperation, TypeError, ValueError):
            return self._invalid_value(value, self.CODE_INVALID, exc_info=True)

        # Decimal's constructor also accepts values such as 'NaN' or
        # '+Inf', which aren't valid in this context.
        # :see: decimal.Decimal._parser
        if not d.is_finite():
            return self._invalid_value(
                value       = value,
                reason      = self.CODE_NON_FINITE,
                exc_info    = True,
            )

        if self.max_precision is not None:
            d = d.quantize(self.max_precision)

        return d


class Int(BaseFilter):
    """
    Interprets the value as an int.

    Strings and other compatible values will be converted, but floats
    will be treated as INVALID.

    Note that Python handles really, really big int values
    transparently, so you don't need to worry about overflow.

    References:
      - http://stackoverflow.com/a/538583
    """
    CODE_DECIMAL = 'not_int'

    templates = {
        CODE_DECIMAL: 'Integer value expected.',
    }

    def _apply(self, value):
        decimal = self._filter(value, Decimal) # type: DecimalType

        if self._has_errors:
            return None

        # Do not allow floats.
        # http://stackoverflow.com/a/19965088
        if decimal % 1:
            return self._invalid_value(value, self.CODE_DECIMAL)

        # Once we get to this point, we're pretty confident that we've
        # got something that can be converted into an int.
        # noinspection PyTypeChecker
        return int(decimal)


@python_2_unicode_compatible
class Max(BaseFilter):
    """
    Enforces a maximum value.

    Note:  Technically, this filter can operate on any type that
    supports comparison, but it tends to be used exclusively with
    numeric types.
    """
    CODE_TOO_BIG = 'too_big'

    templates = {
        CODE_TOO_BIG: 'Value is too large (must be {operator} {max}).',
    }

    def __init__(self, max_value, exclusive=False):
        # type: (Any, bool) -> None
        """
        :param max_value:
            The max value that the Filter will accept.

        :param exclusive:
            Whether to exclude the max value itself as a valid value:

            - True: The incoming value must be _less than_ the max value.
            - False (default): The incoming value must be _less than
              or equal to_ the max value.
        """
        super(Max, self).__init__()

        self.max_value  = max_value
        self.exclusive  = exclusive

    def __str__(self):
        return (
            '{type}({max_value!r}, exclusive={exclusive!r})'.format(
                type        = type(self).__name__,
                max_value   = self.max_value,
                exclusive   = self.exclusive,
            )
        )

    def _apply(self, value):
        # Note that this will yield weird results for string values.
        # We could add better unicode support, if we ever need it.
        # http://stackoverflow.com/q/1097908
        if (
                (value > self.max_value)
            or  (self.exclusive and (value == self.max_value))
        ):
            return self._invalid_value(
                value   = value,
                reason  = self.CODE_TOO_BIG,

                # This only makes sense if `self.exclusive` is False.
                # Better to be consistent and replace all invalid
                # values with `None`.
                # replacement = self.max_value,

                template_vars = {
                    'operator': '<' if self.exclusive else '<=',
                    'max':      self.max_value,
                },
            )

        return value


@python_2_unicode_compatible
class Min(BaseFilter):
    """
    Enforces a minimum value.

    Note:  Technically, this filter can operate on any type that
    supports comparison, but it tends to be used exclusively with
    numeric types.
    """
    CODE_TOO_SMALL  = 'too_small'

    templates = {
        CODE_TOO_SMALL: 'Value is too small (must be {operator} {min}).',
    }

    def __init__(self, min_value, exclusive=False):
        # type: (Any, bool) -> None
        """
        :param min_value:
            The min value that the Filter will accept.

        :param exclusive:
            Whether to exclude the min value itself as a valid value:

            - True: The incoming value must be _greater than_ the min
              value.
            - False (default): The incoming value must be _greater than
              or equal to_ the min value.
        """
        super(Min, self).__init__()

        self.min_value  = min_value
        self.exclusive  = exclusive

    def __str__(self):
        return (
            '{type}({min_value!r}, exclusive={exclusive!r})'.format(
                type        = type(self).__name__,
                min_value   = self.min_value,
                exclusive   = self.exclusive,
            )
        )

    def _apply(self, value):
        # Note that this will yield weird results for string values.
        # We could add better unicode support, if we ever need it.
        # http://stackoverflow.com/q/1097908
        if (
                (value < self.min_value)
            or  (self.exclusive and (value == self.min_value))
        ):
            return self._invalid_value(
                value   = value,
                reason  = self.CODE_TOO_SMALL,

                # This only makes sense if ``self.exclusive`` is False.
                # Better to be consistent and replace all invalid
                # values with ``None``.
                # replacement = self.min_value,

                template_vars = {
                    'operator': '>' if self.exclusive else '>=',
                    'min':      self.min_value,
                },
            )

        return value


class Round(BaseFilter):
    """
    Rounds incoming values to whole numbers or decimals.
    """
    def __init__(self,
            to_nearest  = 1,
            rounding    = ROUND_HALF_UP,
            result_type = DecimalType,
    ):
        # type: (Union[int, Text, DecimalType], Text, type) -> None
        """
        :param to_nearest:
            The value that the filter should round to.

            E.g., ``Round(1)`` rounds to the nearest whole number.

            If you want to round to a float value, it is recommended
            that you provide it as a string or Decimal, to avoid
            floating point problems.

        :param rounding:
            Controls how to round values.

        :param result_type:
            The type of result to return.
        """
        super(Round, self).__init__()

        self.to_nearest = DecimalType(to_nearest)

        # Rounding to negative values isn't supported.
        # I'm not even sure if that concept is valid.
        Min(DecimalType('0')).apply(self.to_nearest)

        self.result_type    = result_type
        self.rounding       = rounding

    def _apply(self, value):
        value = self._filter(value, Decimal) # type: DecimalType

        if self._has_errors:
            return None

        one = DecimalType('1')

        # Scale, round, unscale.
        # Note that we use :py:meth:`decimal.Decimal.quantize` instead
        # of :py:func:`round` to avoid floating-point precision errors.
        # http://stackoverflow.com/a/4340355
        return self.result_type(
                (value * one / self.to_nearest)
                    .quantize(one, rounding=self.rounding)

            *   self.to_nearest
        )



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import json
from datetime import date, datetime, time, tzinfo
from typing import Hashable, Iterable, Mapping, Optional, Sequence, Sized, Text, \
    Union

from dateutil.parser import parse as dateutil_parse
from dateutil.tz import tzoffset
from pytz import utc
from six import binary_type, python_2_unicode_compatible, text_type

from filters.base import BaseFilter, Type
from filters.number import Int, Max, Min

__all__ = [
    'Array',
    'ByteArray',
    'Choice',
    'Date',
    'Datetime',
    'Empty',
    'Length',
    'MaxLength',
    'MinLength',
    'NoOp',
    'NotEmpty',
    'Optional',
    'Required',
]


class Array(Type):
    """
    Validates that the incoming value is a non-string sequence.
    """
    def __init__(self, aliases=None):
        # type: (Optional[Mapping[type, Text]]) -> None
        super(Array, self).__init__(Sequence, True, aliases)

    def _apply(self, value):
        value = super(Array, self)._apply(value) # type: Sequence

        if self._has_errors:
            return None

        if isinstance(value, (binary_type, text_type)):
            return self._invalid_value(
                value   = value,
                reason  = self.CODE_WRONG_TYPE,

                template_vars = {
                    'incoming': self.get_type_name(type(value)),
                    'allowed':  self.get_allowed_type_names(),
                },
            )

        return value


class ByteArray(BaseFilter):
    """
    Converts an incoming value into a bytearray.
    """
    CODE_BAD_ENCODING = 'bad_encoding'

    templates = {
        CODE_BAD_ENCODING:
            'Unable to encode this value using {encoding}.',
    }

    def __init__(self, encoding='utf-8'):
        # type: (Text) -> None
        """
        :param encoding:
            The encoding to use when decoding strings into bytes.
        """
        super(ByteArray, self).__init__()

        self.encoding = encoding

    def _apply(self, value):
        value = self._filter(value, Type(Iterable))

        if self._has_errors:
            return None

        if isinstance(value, bytearray):
            return value

        if isinstance(value, binary_type):
            return bytearray(value)

        if isinstance(value, text_type):
            try:
                return bytearray(value, encoding=self.encoding)
            except UnicodeEncodeError:
                return self._invalid_value(
                    value   = value,
                    reason  = self.CODE_BAD_ENCODING,

                    template_vars = {
                        'encoding': self.encoding,
                    },
                )

        from filters.complex import FilterRepeater
        filtered = self._filter(value, FilterRepeater(
                # Only allow ints and booleans.
                Type(int)
                  # Convert booleans to int (Min and Max require an
                  # exact type match).
                | Int
                  # Min value for each byte is 2^0-1.
                | Min(0)
                  # Max value for each byte is 2^8-1.
                | Max(255)
        ))

        if self._has_errors:
            return None

        return bytearray(filtered)


@python_2_unicode_compatible
class Choice(BaseFilter):
    """
    Expects the value to match one of the items in a set.

    Note:  When matching string values, the comparison is case-
    sensitive!  Use the :py:class:`CaseFold` Filter if you want to
    perform a case-insensitive comparison.
    """
    CODE_INVALID = 'not_valid_choice'

    templates = {
        CODE_INVALID: 'Valid options are: {choices}',
    }

    def __init__(self, choices):
        # type: (Iterable[Hashable]) -> None
        super(Choice, self).__init__()

        self.choices = set(choices)

    def __str__(self):
        return '{type}({choices!r})'.format(
            type = type(self).__name__,

            # Use JSON to mask Python syntax (e.g., remove "u" prefix
            # on unicode strings in Python 2).
            # :py:meth:`Type.__init__`
            choices = json.dumps(sorted(self.choices)),
        )

    def _apply(self, value):
        if value not in self.choices:
            return self._invalid_value(
                value       = value,
                reason      = self.CODE_INVALID,
                exc_info    = True,

                template_vars = {
                    'choices': sorted(self.choices),
                },
            )

        return value


@python_2_unicode_compatible
class Datetime(BaseFilter):
    """
    Interprets the value as a UTC datetime.
    """
    CODE_INVALID = 'not_datetime'

    templates = {
        CODE_INVALID:
            'This value does not appear to be a datetime.',
    }

    def __init__(self, timezone=None, naive=False):
        # type: (Optional[Union[tzinfo, int, float]], bool) -> None
        """
        :param timezone:
            Specifies the timezone to use when the *incoming* value is
            a naive timestamp.  Has no effect on timezone-aware
            timestamps.

            IMPORTANT:  The result is always converted to UTC,
            regardless of the value of the ``timezone`` param!

            You can provide an int/float here, which is the offset from
            UTC in hours (e.g., 5 = UTC+5).

        :param naive:
            If True, the filter will *return* naive datetime objects
            (sans tzinfo).  This is useful e.g., for datetimes that
            will be stored in a database that doesn't understand aware
            timestamps.

            IMPORTANT:  Incoming values are still converted to UTC
            before stripping tzinfo!
        """
        super(Datetime, self).__init__()

        if not isinstance(timezone, tzinfo):
            if timezone in [0, None]:
                timezone = utc
            else:
                # Assume that we got an int/float instead.
                timezone = tzoffset(
                    name    = 'UTC{offset:+}'.format(offset=timezone),
                    offset  = float(timezone) * 3600.0,
                )

        self.timezone   = timezone
        self.naive      = naive

    def __str__(self):
        return '{type}(timezone={timezone!r}, naive={naive!r})'.format(
            type        = type(self).__name__,
            timezone    = self.timezone,
            naive       = self.naive,
        )

    def _apply(self, value):
        if isinstance(value, datetime):
            parsed = value
        elif isinstance(value, date):
            # http://stackoverflow.com/a/1937636
            parsed = datetime.combine(value, time.min)
        else:
            try:
                #
                # It's a shame we can't pass ``tzinfos`` to
                # :py:meth:`dateutil_parse.parse`; ``tzinfos`` only has
                # effect if we also specify ``ignoretz = True``, which
                # we definitely don't want to do here!
                #
                # https://dateutil.readthedocs.org/en/latest/parser.html#dateutil.parser.parse
                #
                parsed = dateutil_parse(value)
            except ValueError:
                return self._invalid_value(
                    value       = value,
                    reason      = self.CODE_INVALID,
                    exc_info    = True,
                )

        if not parsed.tzinfo:
            parsed = parsed.replace(tzinfo=self.timezone)

        # Always covert to UTC.
        aware_result = parsed.astimezone(utc)

        return (
            aware_result.replace(tzinfo=None)
                if self.naive
                else aware_result
        )


class Date(Datetime):
    """
    Interprets the value as a UTC date.

    Note that the value is first converted to a datetime with UTC
    timezone, which may cause the resulting date to appear to be
    off by +/- 1 day (does not apply if the value is already a date
    object).
    """
    CODE_INVALID = 'not_date'

    templates = {
        CODE_INVALID: 'This value does not appear to be a date.',
    }

    def _apply(self, value):
        if isinstance(value, date) and not isinstance(value, datetime):
            return value

        filtered = super(Date, self)._apply(value) # type: datetime

        # Normally we return `None` if we get any errors, but in this
        # case, we'll let the superclass method decide.
        return filtered if self._has_errors else filtered.date()


class Empty(BaseFilter):
    """
    Expects the value to be empty.

    In this context, "empty" is defined as having zero length.  Note
    that this Filter considers values that do not have length to be
    not empty (in particular, False and 0 are not considered empty
    here).
    """
    CODE_NOT_EMPTY = 'not_empty'

    templates = {
        CODE_NOT_EMPTY: 'Empty value expected.',
    }

    def _apply(self, value):
        try:
            length = len(value)
        except TypeError:
            length = 1

        return (
            self._invalid_value(value, self.CODE_NOT_EMPTY)
                if length
                else value
        )


@python_2_unicode_compatible
class Length(BaseFilter):
    """
    Ensures incoming values have exactly the right length.
    """
    CODE_TOO_LONG   = 'too_long'
    CODE_TOO_SHORT  = 'too_short'

    templates = {
        CODE_TOO_LONG:
            'Value is too long (length must be exactly {expected}).',
        CODE_TOO_SHORT:
            'Value is too short (length must be exactly {expected}).',
    }

    def __init__(self, length):
        # type: (int) -> None
        super(Length, self).__init__()

        self.length = length

    def __str__(self):
        return '{type}(length={length!r})'.format(
            type    = type(self).__name__,
            length  = self.length,
        )

    def _apply(self, value):
        value = self._filter(value, Type(Sized))

        if self._has_errors:
            return None

        if len(value) > self.length:
            return self._invalid_value(
                value   = value,
                reason  = self.CODE_TOO_LONG,

                template_vars = {
                    'expected': self.length,
                },
            )
        elif len(value) < self.length:
            return self._invalid_value(
                value   = value,
                reason  = self.CODE_TOO_SHORT,

                template_vars = {
                    'expected': self.length,
                },
            )

        return value


@python_2_unicode_compatible
class MaxLength(BaseFilter):
    """
    Enforces a maximum length on the value.
    """
    CODE_TOO_LONG = 'too_long'

    templates = {
        CODE_TOO_LONG: 'Value is too long (length must be < {max}).',
    }

    def __init__(self, max_length):
        # type: (int) -> None
        super(MaxLength, self).__init__()

        self.max_length = max_length

    def __str__(self):
        return '{type}({max_length!r})'.format(
            type        = type(self).__name__,
            max_length  = self.max_length,
        )

    def _apply(self, value):
        if len(value) > self.max_length:
            # Note that we do not truncate the value:
            #   - It's not always clear which end we should truncate
            #     from.
            #   - We should keep this filter's behavior consistent with
            #     that of MinLength.
            return self._invalid_value(
                value   = value,
                reason  = self.CODE_TOO_LONG,

                template_vars = {
                    'length':   len(value),
                    'max':      self.max_length,
                },
            )

        return value


class MinLength(BaseFilter):
    """
    Enforces a minimum length on the value.
    """
    CODE_TOO_SHORT = 'too_short'

    templates = {
        CODE_TOO_SHORT: 'Value is too short (length must be > {min}).',
    }

    def __init__(self, min_length):
        # type: (int) -> None
        super(MinLength, self).__init__()

        self.min_length = min_length

    def __str__(self):
        return '{type}({min_length!r})'.format(
            type        = type(self).__name__,
            min_length  = self.min_length,
        )

    def _apply(self, value):
        if len(value) < self.min_length:
            #
            # Note that we do not pad the value:
            #   - It is not clear to which end(s) we should add the
            #     padding.
            #   - It is not clear what the padding value(s) should be.
            #   - We should keep this filter's behavior consistent with
            #     that of MaxLength.
            #
            return self._invalid_value(
                value   = value,
                reason  = self.CODE_TOO_SHORT,

                template_vars = {
                    'length':       len(value),
                    'min':          self.min_length,
                },
            )

        return value


class NoOp(BaseFilter):
    """
    Filter that does nothing, used when you need a placeholder Filter
    in a FilterChain.
    """
    def _apply(self, value):
        return value


@python_2_unicode_compatible
class NotEmpty(BaseFilter):
    """
    Expects the value not to be empty.

    In this context, "empty" is defined as having zero length.  Note
    that this filter considers values that do not have length to be
    not empty (in particular, False and 0 are not considered empty
    here).

    By default, this filter treats ``None`` as valid, just like every
    other filter.  However, you can configure the filter to reject
    ``None`` in its initializer method.
    """
    CODE_EMPTY = 'empty'

    templates = {
        CODE_EMPTY: 'Non-empty value expected.',
    }

    def __init__(self, allow_none=True):
        # type: (bool) -> None
        """
        :param allow_none:
            Whether to allow ``None``.
        """
        super(NotEmpty, self).__init__()

        self.allow_none = allow_none

    def __str__(self):
        return '{type}(allow_none={allow_none!r})'.format(
            type        = type(self).__name__,
            allow_none  = self.allow_none,
        )

    def _apply(self, value):
        try:
            length = len(value)
        except TypeError:
            length = 1

        return value if length else self._invalid_value(value, self.CODE_EMPTY)

    def _apply_none(self):
        if not self.allow_none:
            return self._invalid_value(None, self.CODE_EMPTY)

        return None


class Required(NotEmpty):
    """
    Same as NotEmpty, but with ``allow_none`` hard-wired to ``False``.

    This filter is the only exception to the "``None`` passes by
    default" rule.
    """
    templates = {
        NotEmpty.CODE_EMPTY: 'This value is required.',
    }

    def __init__(self):
        super(Required, self).__init__(allow_none=False)


@python_2_unicode_compatible
class Optional(BaseFilter):
    """
    Changes empty and null values into a default value.

    In this context, "empty" is defined as having zero length.  Note
    that this Filter considers values that do not have length to be
    not empty (in particular, False and 0 are not considered empty
    here).
    """
    def __init__(self, default=None):
        """
        :param default:
            The default value used to replace empty values.
        """
        super(Optional, self).__init__()

        self.default = default

    def __str__(self):
        return '{type}(default={default!r})'.format(
            type    = type(self).__name__,
            default = self.default,
        )

    def _apply(self, value):
        try:
            length = len(value)
        except TypeError:
            length = 1

        return value if length > 0 else self.default

    def _apply_none(self):
        return self.default
